import unittest
from zope.testing import doctest
from zope.component import testing
from Testing.ZopeTestCase import FunctionalDocFileSuite
from Products.PloneTestCase import PloneTestCase
from Products.PloneTestCase.layer import onsetup, PloneSite
from Products.Five import zcml

@onsetup
def load_zcml():
    import p4a.subtyper
    zcml.load_config('configure.zcml', p4a.subtyper)

load_zcml()
PloneTestCase.setupPloneSite()

def test_suite():
    flags = doctest.NORMALIZE_WHITESPACE|doctest.ELLIPSIS

    unitsuite = unittest.TestSuite((
        doctest.DocFileSuite('subtyping.txt',
                             setUp=testing.setUp,
                             tearDown=testing.tearDown,
                             optionflags=flags),
        ))

    fsuite = unittest.TestSuite((
        FunctionalDocFileSuite('browser.txt',
                               package='p4a.subtyper',
                               optionflags=flags,
                               test_class=PloneTestCase.FunctionalTestCase),
        ))
    fsuite.layer = PloneSite

    return unittest.TestSuite((unitsuite, fsuite))



from zope import component
from p4a.subtyper.interfaces import ISubtyper

from Products.CMFCore import DirectoryView
DirectoryView.registerDirectory('skins', globals())
from zope.i18nmessageid import MessageFactory
subtyperMessageFactory = MessageFactory('p4a.subtyper')

class activated(property):
    """A descriptor for setting or getting whether a subtype
    has been applied to an object.
    """

    def __init__(self, desc_name, attr=None):
        self.desc_name = desc_name
        self.attr = attr

    def get_obj(self, obj):
        if not self.attr:
            return obj
        return getattr(obj, self.attr, None)

    def __get__(self, obj, type=None):
        subtyper = component.queryUtility(ISubtyper)
        if subtyper is None:
            return False
        realobj = self.get_obj(obj)
        if realobj is None:
            return False

        descwithname = subtyper.existing_type(realobj)
        if descwithname is None:
            return False
        return descwithname.name == self.desc_name

    def __set__(self, obj, v):
        subtyper = component.getUtility(ISubtyper)
        v = bool(v)
        orig = self.__get__(obj)
        realobj = self.get_obj(obj)
        if v and not orig:
            subtyper.change_type(realobj, self.desc_name)
        elif not v and orig:
            subtyper.remove_type(realobj)

    def __delete__(self, obj):
        self.__set__(obj, False)



from Products.CMFCore.utils import getToolByName
from plone.app.contentmenu.interfaces import IActionsSubMenuItem

from zope import interface
from zope.app.publisher.browser.menu import BrowserSubMenuItem
from zope.component import getMultiAdapter

from p4a.subtyper import subtyperMessageFactory as _

class SubtypesSubMenuItem(BrowserSubMenuItem):
    interface.implements(IActionsSubMenuItem)

    title = _(u'Sub-types')
    description = u''
    submenuId = u'subtypes'
    order = 9
    extra = {'id': 'subtypes'}

    @property
    def action(self):
        return self.context.absolute_url()+ '/subtypes'

    def available(self):
        """decide wether or not to show the subtyper-menu"""

        # don't show the menu if the item doesn't have any subtypes
        view = getMultiAdapter((self.context, self.request), name=u'subtyper')
        if not view.has_possible_types():
            return False

        # We'll show the menu only to managers,
        # see also the permission in p4a.subtyper's configure.zcml
        # TODO: p4a should set it own permissions
        mtool = getToolByName(self.context, 'portal_membership')
        if mtool.checkPermission('Manage portal', self.context):
            return True
        else:
            return False



from Acquisition import aq_inner
from zope.interface import implements, implementer, Interface
#import zope.component
from zope.component import getUtilitiesFor, adapter

import Products.Archetypes.interfaces
from p4a.subtyper import interfaces


class _DescriptorsMixin(object):

    def __init__(self, possible=[], comment=None):
        self._possible = possible
        self._comment = comment

    @property
    def possible(self):
        return self._possible

    def __str__(self):
        return '<PossibleDescriptors comment=%s>' % (self._comment or '')
    __repr__ = __str__


class PossibleDescriptors(_DescriptorsMixin):
    implements(interfaces.IPossibleDescriptors)


class PortalTypedPossibleDescriptors(_DescriptorsMixin):
    implements(interfaces.IPortalTypedPossibleDescriptors)


@adapter(Products.Archetypes.interfaces.IBaseFolder)
@implementer(interfaces.IPossibleDescriptors)
def folderish_possible_descriptors(context):
    possible = getUtilitiesFor(\
               interfaces.IFolderishContentTypeDescriptor)
    possible = [(n, c) for n, c in possible]
    return PossibleDescriptors(possible, 'folderish')


@adapter(Interface)
@implementer(interfaces.IPossibleDescriptors)
def nonfolderish_possible_descriptors(context):
    portal_type = getattr(aq_inner(context), 'portal_type', None)
    if portal_type is None:
        return PossibleDescriptors()

    all = getUtilitiesFor(\
          interfaces.IContentTypeDescriptor)
    all = set([(n, c) for n, c in all])
    folderish = getUtilitiesFor(\
          interfaces.IFolderishContentTypeDescriptor)
    folderish = set([(n, c) for n, c in folderish])

    return PossibleDescriptors(list(all.difference(folderish)),
                               'nonfolderish')


@adapter(Products.Archetypes.interfaces.IBaseFolder)
@implementer(interfaces.IPortalTypedPossibleDescriptors)
def portal_typed_folderish_possible_descriptors(context):
    portal_type = getattr(aq_inner(context), 'portal_type', None)
    if portal_type is None:
        return PortalTypedPossibleDescriptors()

    possible = getUtilitiesFor(\
               interfaces.IPortalTypedFolderishDescriptor)
    return PortalTypedPossibleDescriptors([(n, c) for n, c in possible
                                           if c.for_portal_type == portal_type],
                                           'portal_typed_folderish')


@adapter(Interface)
@implementer(interfaces.IPortalTypedPossibleDescriptors)
def portal_typed_nonfolderish_possible_descriptors(context):
    portal_type = getattr(aq_inner(context), 'portal_type', None)
    if portal_type is None:
        return PortalTypedPossibleDescriptors()

    all = getUtilitiesFor(\
          interfaces.IPortalTypedDescriptor)
    folderish = getUtilitiesFor(\
          interfaces.IPortalTypedFolderishDescriptor)

    all = set([(n, c) for n, c in all if c.for_portal_type == portal_type])
    folderish = set([(n, c) for n, c in folderish
                     if c.for_portal_type == portal_type])

    return PortalTypedPossibleDescriptors(list(all.difference(folderish)),
                               'portal_typed_nonfolderish')



from persistent.dict import PersistentDict

from zope.component import queryUtility
from zope.component import getUtility

from zope.event import notify

import zope.interface
#from zope.interface import directlyProvides
#from zope.interface import directlyProvidedBy
from zope.interface import alsoProvides
from zope.interface import implements

try:
    from zope.annotation import interfaces as annoifaces
except ImportError, err:
    from zope.app.annotation import interfaces as annoifaces
from p4a.subtyper import interfaces


class SubtypeEvent(object):
    __doc__ = interfaces.ISubtypeEvent
    implements(interfaces.ISubtypeEvent)

    def __init__(self, object, subtype):
        self.object = object
        self.subtype = subtype


class SubtypeAddedEvent(SubtypeEvent):
    __doc__ = interfaces.ISubtypeAddedEvent
    implements(interfaces.ISubtypeAddedEvent)


class SubtypeRemovedEvent(SubtypeEvent):
    __doc__ = interfaces.ISubtypeRemovedEvent
    implements(interfaces.ISubtypeRemovedEvent)


class SubtypeError(Exception):
    pass


class NoSubtypeDefined(SubtypeError):
    pass


class ExistingSubtypeDefined(SubtypeError):
    pass


class DescriptorWithName(object):
    implements(interfaces.IDescriptorWithName)

    def __init__(self, name, descriptor):
        self.name = name
        self.descriptor = descriptor

    def __str__(self):
        return '<DescriptorWithName name=%s; descriptor=%s>' % \
               (self.name, str(self.descriptor))
    __repr__ = __str__


class _DescriptorInfo(object):
    ANNO_KEY = 'p4a.subtyper.DescriptorInfo'

    def __init__(self, context):
        self.context = context
        self.anno = annoifaces.IAnnotations(context, None)

    def _ensure_info(self):
        if self.anno is None:
            annoifaces.IAnnotations(self.context)
        d = self.anno.get(self.ANNO_KEY, None)
        if d is None:
            d = PersistentDict()
            self.anno[self.ANNO_KEY] = d
        return d

    def _set_name(self, v):
        info = self._ensure_info()
        info['descriptor_name'] = v

    def _get_name(self):
        if self.anno is None:
            return None
        d = self.anno.get(self.ANNO_KEY, None)
        if d is None:
            return None
        return d.get('descriptor_name', None)
    name = property(_get_name, _set_name)


class Subtyper(object):
    __doc__ = interfaces.ISubtyper.__doc__
    implements(interfaces.ISubtyper)

    def possible_types(self, obj):
        possible = interfaces.IPossibleDescriptors(obj)
        return (DescriptorWithName(n, c) for n, c in possible.possible)

    def _remove_type(self, obj):
        info = _DescriptorInfo(obj)
        name = info.name
        descriptor = None

        if name is not None:
            info.name = None
            descriptor = queryUtility( \
                interfaces.IContentTypeDescriptor, name=name)
            if descriptor is not None:
                directlyProvides = zope.interface.directlyProvides
                directlyProvidedBy = zope.interface.directlyProvidedBy
                should_have = directlyProvidedBy(obj) - \
                              descriptor.type_interface -\
                              interfaces.ISubtyped
                directlyProvides(obj, should_have)

        return descriptor

    def remove_type(self, obj):
        descriptor = self._remove_type(obj)
        if descriptor is None:
            raise NoSubtypeDefined()
        notify(SubtypeRemovedEvent(obj, descriptor))

    def _add_type(self, obj, descriptor_name):
        info = _DescriptorInfo(obj)
        if info.name is not None:
            raise ExistingSubtypeDefined(str(info.name))
        descriptor = getUtility( \
            interfaces.IContentTypeDescriptor, name=descriptor_name)
        info.name = descriptor_name
        alsoProvides(obj, (interfaces.ISubtyped,
                                          descriptor.type_interface, ))
        return descriptor

    def change_type(self, obj, descriptor_name):
        removed = self._remove_type(obj)
        notify(SubtypeRemovedEvent(obj, removed))

        added = self._add_type(obj, descriptor_name)
        notify(SubtypeAddedEvent(obj, added))

    def existing_type(self, obj):
        info = _DescriptorInfo(obj)
        if info.name != None:
            c = queryUtility( \
                interfaces.IContentTypeDescriptor, name=info.name)
            return DescriptorWithName(info.name, c)
        return None

    def get_named_type(self, name):
        return getUtility(interfaces.IContentTypeDescriptor,
                                         name=name)



from Acquisition import aq_base
from zope.interface import Interface
from zope.interface import implements
from zope.component import getUtility

from Products.Five.browser import BrowserView
from Products.statusmessages.interfaces import IStatusMessage
from Products.CMFDynamicViewFTI import interfaces as IDynamicallyViewableInterface

from p4a.subtyper import interfaces
#from p4a.subtyper import utils
from p4a.subtyper import subtyperMessageFactory as _

class ISubtyperView(Interface):

    def possible_types():
        pass

    def has_possible_types():
        pass

    def change_type():
        pass


class SubtyperView(BrowserView):
    """View for introspecting and possibly changing subtype info for the
    current context.
    """

    implements(ISubtyperView)

    def possible_types(self):
        subtyper = getUtility(interfaces.ISubtyper)
        return subtyper.possible_types(self.context)

    def has_possible_types(self):
        types = [subtype for subtype in self.possible_types()]
        return len(types) > 0

    def _redirect(self, msg):
        url = self.context.absolute_url()
        if hasattr(aq_base(self.context), 'getLayout'):
            layout = self.context.getLayout() or ''
            if layout:
                if not url.endswith('/'):
                    url += '/'
                url += layout
        IStatusMessage(self.request).addStatusMessage(msg, type='info')
        self.request.response.redirect(url)
        return ''

    def change_type(self):
        """Change the sub type of the current context.
        """

        subtyper = getUtility(interfaces.ISubtyper)

        subtype_name = self.request.get('subtype', None)
        if subtype_name:
            existing = subtyper.existing_type(self.context)
            subtype = subtyper.get_named_type(subtype_name)
            if existing is not None and existing.name == subtype_name:
                selected_layout = False
                dynamic_view = IDynamicallyViewableInterface.IDynamicallyViewable(self.context, None)
                if dynamic_view is not None:
                    if self.context.getLayout() in dynamic_view.getAvailableViewMethods():
                        selected_layout = True
                subtyper.remove_type(self.context)
                msg = _(u'Removed ${title} subtype', mapping={'title': subtype.title})
                # Check if the default view has disappeared:
                if selected_layout:
                    dynamic_view = IDynamicallyViewableInterface.IDynamicallyViewable(self.context, None)
                    if dynamic_view is None or (self.context.getLayout() in
                       dynamic_view.getAvailableViewMethods()):
                        if self.context.hasProperty('layout'):
                            self.context.manage_delProperties(['layout'])
            else:
                subtyper.change_type(self.context, subtype_name)
                msg = _(u'Changed subtype to ${title}', mapping={'title': subtype.title})
        else:
            msg = _(u'No subtype specified')

        return self._redirect(msg)



from p4a.common import site
from p4a.z2utils import utils
from Products.CMFCore.utils import getToolByName
from Products.CMFCore import utils as cmfutils
from Products.CMFCore import DirectoryView
from p4a.subtyper import interfaces
from zope.interface import Interface
from plone.app.contentmenu.interfaces import IContentMenuItem
    
import logging
logger = logging.getLogger('p4a.subtyper.sitesetup')

try:
    import five.localsitemanager
    HAS_FLSM = True
    logger.info('Using five.localsitemanager')
except ImportError, err:
    HAS_FLSM = False


# Unused:
def setup_portal(portal):
    site.ensure_site(portal)
    setup_site(portal)


# Unused:
def setup_site(site):
    # In 2.5, install the subtyper profile:
    mt = cmfutils.getToolByName(site, 'portal_migration')
    plone_version = mt.getInstanceVersion()

    # Register profile for Plone 3.
    if int(plone_version[0]) >= 3:
        quickinstaller_tool = getToolByName(site, 'portal_quickinstaller')
        quickinstaller_tool.installProduct('p4a.subtyper')

    if plone_version[0:3] == '2.5':
        # Setup only needed for Plone 3.0
        skin_tool = cmfutils.getToolByName(site, 'portal_skins')
        path = None
        for path in DirectoryView.manage_listAvailableDirectories():
            if path.endswith('p4a_subtyper'):
                break
        assert(path is None, "Subtyper skin directory not found")
        if 'p4a_subtyper' not in skin_tool.objectIds():
            DirectoryView.createDirectoryView(skin_tool, path)

        for skin_name, paths in skin_tool.getSkinPaths():
            if not 'p4a_subtyper' in paths:
                paths = paths.split(',')
                index = paths.index('plone_templates')
                paths.insert(index, 'p4a_subtyper')
                paths = ','.join(paths)
                skin_tool._getSelections()[skin_name] = paths

# Used:
def unsetup_portal(portal, reinstall=False, reindex=True):
    if reinstall:
        # Do nothing.
        return
    if reindex:
        # Setting marker interfaces doesn't reindex objects, so we need to
        # reindex object_provides to make sure it's up to date. If called
        # from other products that already have done this pass reindex=False.
        portal.portal_catalog.manage_reindexIndex(('object_provides', ))
    # Then we can use the removal utility to unregister all of them:
    count = utils.remove_marker_ifaces(portal, interfaces.ISubtyped)
    logger.warn('Removed ISubtyped interface from %i objects for '
                'cleanup' % count)

    sm = portal.getSiteManager()
    if sm.adapters.lookup((Interface, Interface), IContentMenuItem, u'subtypes') is not None:
        sm.unregisterAdapter(None, (Interface, Interface), IContentMenuItem, u'subtypes')




import unittest
from zope.testing import doctest
from Testing.ZopeTestCase import FunctionalDocFileSuite
import Products.PloneTestCase.layer

from Products.PloneTestCase import PloneTestCase
try:
    import Products.Plone4ArtistsSubtyper
    PloneTestCase.installProduct('Plone4ArtistsSubtyper')
except ImportError, err:
    # this is ok
    pass

PloneTestCase.setupPloneSite()


def test_suite():
    flags = doctest.NORMALIZE_WHITESPACE|doctest.ELLIPSIS
    fsuite = unittest.TestSuite(())
    try:
        # this is to make sure the tests only run
        # in a plone 2.5 / zope 2.9 environment
        import zope.app.annotation.tests
        fsuite.append(FunctionalDocFileSuite('browser.txt',
                                             package='p4a.subtyper.contentmenu',
                                             optionflags=flags), )
        fsuite.layer = Products.PloneTestCase.layer.PloneSite
    except ImportError, err:
        # this is ok
        pass

    return fsuite



from Products.Five.browser import BrowserView
import Globals
import AccessControl.SecurityInfo


class ContentMenuView(BrowserView):
    """A view for the content menu.
    """

    security = AccessControl.SecurityInfo.ClassSecurityInfo()

    security.declarePublic('macros')

    @property
    def macros(self):
        return self.index.macros

Globals.InitializeClass(ContentMenuView)


class TestView(BrowserView):
    """Simple replacement of the standard zope2 test function.
    """

    def test(self, condition, trueval, falseval):
        if condition:
            return trueval
        return falseval

    def __call__(self, condition, trueval, falseval):
        return self.test(condition, trueval, falseval)



import os
from setuptools import setup, find_packages


def read(*rnames):
    return open(os.path.join(os.path.dirname(__file__), *rnames)).read()

version = '1.2.2'

long_description = (
    read('README.txt')
    + '\n' +
    read('CHANGES.txt')
    )


setup(name='p4a.subtyper',
      version=version,
      description="Subtyping framework for Plone",
      long_description=long_description,
      classifiers=[
        "Framework :: Zope2",
        "Framework :: Zope3",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
        ],
      keywords='',
      author='Rocky Burt',
      author_email='rocky@serverzen.com',
      #url='http://www.plone4artists.org/svn/projects/p4a.subtyper/',
      url='https://svn.plone.org/svn/collective/p4a/p4a.subtyper/',
      license='GPL',
      packages=find_packages(exclude=['ez_setup']),
      namespace_packages=['p4a'],
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          'setuptools',
          'p4a.z2utils'
      ],
      entry_points="""
      # -*- Entry points: -*-
      """,
      )



from zope.interface import Interface, Attribute
from zope.schema import List, Object, Set, TextLine, Text, InterfaceField


class ISubtyped(Interface):
    """An object that has been subtyped with this machinery should provide
    this interface.
    """


class IContentTypeDescriptor(Interface):
    """A descriptor representing information about a content type.
    """

    title = TextLine(title=u'Title')
    description = Text(title=u'Description')
    type_interface = InterfaceField(title=u'Type Interface')


class IFolderishContentTypeDescriptor(IContentTypeDescriptor):
    """A descriptor which allows to describe what possible objects of
    what particular portal_type's can be added as child items.
    """

    allowed_child_portal_types = Set( \
        title=u'Allowed Child Portal Types')


class _IPortalTypedMixin(IContentTypeDescriptor):
    """A set of common fields to use for any descriptor which supports
    discrimination by portal_type.
    """

    for_portal_type = TextLine(title=u'For Portal Type')


class IPortalTypedDescriptor(IContentTypeDescriptor,
                             _IPortalTypedMixin):
    """A descriptor which can be discriminated by portal_type.
    """


class IPortalTypedFolderishDescriptor(IFolderishContentTypeDescriptor,
                                      _IPortalTypedMixin):
    """A folderish descriptor which can be discriminated by portal_type.
    """


class IPossibleDescriptors(Interface):
    """A method of getting the possible descriptors on a given context.
    Intended to be implemented by adapters.
    """

    possible = List( \
        title=u'Possible',
        value_type=Object(title=u'Possible',
                                      schema=IContentTypeDescriptor))


class IPortalTypedPossibleDescriptors(IPossibleDescriptors):
    """A method of getting the possible descriptors on a given context,
    for a given portal_type.
    Intended to be implemented by adapters.
    """


class ISubtyper(Interface):
    """The actual subtyping engine.
    """

    def possible_types(obj):
        """All possible types for a given object.  Will return an iterable
        of objects implementing IContentTypeDescriptor.
        """

    def change_type(obj, descriptor_name):
        """Will change the sub type of the given object to the specified
        type.  If no sub type was already set on the object, the target type
        will be added.  If a sub type already exists, it will be removed
        and the target type will be added.
        """

    def remove_type(obj):
        """Remove any sub type on the object.
        """

    def existing_type(obj):
        """Return the existing type of an object.
        """

    def get_named_type(name):
        """Return descriptor represented by name.
        """


class ISubtypeEvent(Interface):
    object = Attribute('The subject of the event')
    subtype = InterfaceField(title=u'Type Interface')


class ISubtypeAddedEvent(ISubtypeEvent):
    """Fired when a subtype was added to an object.
    """


class ISubtypeRemovedEvent(ISubtypeEvent):
    """Fired when a subtype was removed from an object.
    """


class IDescriptorWithName(Interface):
    """A simple way to match a descriptor with it's registered name.
    """

    name = TextLine(title=u'Name')
    descriptor = Object(title=u'Descriptor', schema=IContentTypeDescriptor)
